from typing import Annotated

from zayt.di.inject import Inject
from zayt.di.service.model import ServiceDependency, ServiceSpec
from zayt.di.service.parse import parse_service_spec


def test_parse_service_spec():
    class ServiceClass:
        dependency: Annotated[str, Inject]

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name=None))],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_with_interface():
    class ServiceClass:
        pass

    result = parse_service_spec(ServiceClass, provides=object)
    assert result == ServiceSpec(
        service=object,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_with_name():
    class ServiceClass:
        pass

    result = parse_service_spec(ServiceClass, name="service")
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name="service",
        scoped=False,
        dependencies=[],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_more_dependencies():
    class ServiceClass:
        dependency: Annotated[str, Inject]
        other: Annotated[int, Inject]

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[
            ("dependency", ServiceDependency(service=str, name=None)),
            ("other", ServiceDependency(service=int, name=None)),
        ],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_optional_dependency():
    class ServiceClass:
        dependency: Annotated[str, Inject] = None

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[
            ("dependency", ServiceDependency(service=str, name=None, optional=True))
        ],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_non_inject_dependency():
    class ServiceClass:
        dependency: Annotated[str, Inject]
        property: str

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name=None))],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_spec_named_dependency():
    class ServiceClass:
        dependency: Annotated[str, Inject("test")]

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name="test"))],
        initializer=None,
        finalizer=None,
    )


def test_parse_service_with_annotated_non_dependency():
    class ServiceClass:
        dependency: Annotated[str, Inject]
        property: Annotated[int, object]

    result = parse_service_spec(ServiceClass)
    assert result == ServiceSpec(
        service=ServiceClass,
        impl=ServiceClass,
        factory=None,
        name=None,
        scoped=False,
        dependencies=[("dependency", ServiceDependency(service=str, name=None))],
        initializer=None,
        finalizer=None,
    )



from typing import Generic, Protocol, TypeVar

import pytest

from zayt.util.base_types import get_base_types

T1 = TypeVar("T1")


class Interface1(Generic[T1]):
    pass


T2 = TypeVar("T2")


class Interface2(Generic[T2], Interface1[int]):
    pass


class Impl1(Interface1[int]):
    pass


class Impl12(Impl1):
    pass


class Impl2(Interface2[str]):
    pass


class Impl22(Impl2):
    pass


class Proto1(Protocol[T1]):
    pass


class ProtoImpl1(Proto1[str]):
    pass


@pytest.mark.parametrize(
    "base_type,expected",
    [
        (Impl12, [Impl12, Impl1, Interface1[int], Interface1]),
        (
            Impl22,
            [Impl22, Impl2, Interface2[str], Interface2, Interface1[int], Interface1],
        ),
        (Interface1[str], [Interface1[str], Interface1]),
        (Interface2[str], [Interface2[str], Interface2, Interface1[int], Interface1]),
        (ProtoImpl1, [ProtoImpl1, Proto1[str], Proto1]),
    ],
    ids=[
        "1 level generic",
        "2 levels generic",
        "1 level interface input",
        "2 levels interface input",
        "Using 'Protocol'",
    ],
)
def test_get_base_types(base_type, expected):
    base_types = get_base_types(base_type)
    assert base_types == expected



from typing import Annotated

import pytest

from zayt.di.container import Container
from zayt.di.error import DependencyLoopError
from zayt.di.inject import Inject


class Service1:
    service2: Annotated["Service2", Inject]


class Service2:
    service1: Annotated[Service1, Inject]


def factory1(service2: Service2) -> Service1:
    instance = Service1()
    instance.service2 = service2
    return instance


def factory2(service1: Service1) -> Service2:
    instance =  Service2()
    instance.service1 = service1
    return instance


async def test_service_class(ioc: Container):
    ioc.register(Service1)
    ioc.register(Service2)

    service1 = await ioc.get(Service1)
    service2 = await ioc.get(Service2)

    assert service1 is service2.service1
    assert service2 is service1.service2


async def test_service_factory(ioc: Container):
    ioc.register(factory1)
    ioc.register(factory2)

    with pytest.raises(DependencyLoopError):
        await ioc.get(Service1)



from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt


async def test_extension(monkeypatch):
    monkeypatch.chdir(__file__.rsplit("/", maxsplit=1)[0])
    settings = Settings(
        default_settings
        | {
            "modules": [f"{__package__}.extension"],
        }
    )

    app = Zayt(settings)

    # pylint: disable=protected-access
    await app._lifespan_startup()

    assert settings.tested



from zayt.util.package_scan import scan_packages
from zayt.web.lifecycle.decorator import ATTRIBUTE_BACKGROUND, ATTRIBUTE_STARTUP


def _predicate_startup_hooks(item) -> bool:
    return getattr(item, ATTRIBUTE_STARTUP, False)


def find_startup_hooks(*modules):
    return list(scan_packages(*modules, predicate=_predicate_startup_hooks))


def _predicate_background_services(item) -> bool:
    return getattr(item, ATTRIBUTE_BACKGROUND, False)


def find_background_services(*modules):
    return list(scan_packages(*modules, predicate=_predicate_background_services))



from typing import Any

from zayt.di.container import Container


class MyInterceptor:
    async def intercept(self, instance: Any, _service_type: type):
        setattr(instance, "intercepted", True)


class MyService:
    pass


async def test_intercept(ioc: Container):
    ioc.register(MyService)
    ioc.interceptor(MyInterceptor)

    instance = await ioc.get(MyService)

    assert getattr(instance, "intercepted")



import json

import pytest

from zayt.conf import env


def test_str(monkeypatch):
    monkeypatch.setenv("TEST", "value")
    result = env("TEST")
    assert result == "value"


@pytest.mark.parametrize(
    "value, typecast, expected",
    [
        ("1", int, 1),
        ("1.2", float, 1.2),
        ("true", bool, True),
        ("1", bool, True),
        ("false", bool, False),
        ("0", bool, False),
        ("any", bool, False),
        ('{"key": "value"}', json, {"key": "value"}),
        ("", int, None),
        ("", float, None),
        ("", str, None),
        ("", bool, None),
    ],
)
def test_cast(value, typecast, expected, monkeypatch):
    monkeypatch.setenv("TEST", value)
    result = env("TEST", cast=typecast)
    assert result == expected


def test_cast_function(monkeypatch):
    monkeypatch.setenv("TEST", "value")

    def cast_func(value: str):
        return [value]

    result = env("TEST", cast=cast_func)
    assert result == ["value"]


def test_cast_class_init(monkeypatch):
    monkeypatch.setenv("TEST", "value")

    class MyType:
        def __init__(self, value):
            self.value = value

    result = env("TEST", cast=MyType)
    assert isinstance(result, MyType)


def test_default():
    result = env("TEST", default="default")
    assert result == "default"


def test_missing_variable_should_fail():
    with pytest.raises(ValueError, match="Environment variable 'TEST' not set"):
        env("TEST")



import asyncio
import multiprocessing as mp
from pathlib import Path

from httpx import ASGITransport, AsyncClient

from .util import with_app


@with_app
async def _run(app):
    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://server"
    ) as client:
        response = await client.get("/", params={"name": "Python", "number": 3})
        assert response.json() == {"greeting": "Hello, Python!", "number": 3}

        response = await client.get("/Asgi")
        assert response.json() == {"greeting": "Hello, Asgi!"}

        response = await client.post("/", json={"greeting": "Hello, World!"})
        assert response.json() == {"result": {"greeting": "Hello, World!"}}

        for path in ["/multiple", "/annotations"]:
            response = await client.get(path)
            assert response.json() == {"path": path}


def run():
    return asyncio.run(_run())


def test_example(monkeypatch):
    path = Path(__file__).parents[2] / "examples" / "hello_world"
    monkeypatch.syspath_prepend(path)
    monkeypatch.chdir(path)

    with mp.Pool(processes=1) as pool:
        pool.apply(run)



from httpx import AsyncClient
from httpx_ws import aconnect_ws, WebSocketDisconnect
from httpx_ws.transport import ASGIWebSocketTransport, UnhandledWebSocketEvent

import pytest

from zayt.conf.defaults import default_settings
from zayt.conf.settings import Settings
from zayt.web.application import Zayt


@pytest.mark.skip
async def test_websocket_not_found():
    settings = Settings(
        default_settings | {"__application__": __package__ + ".application"}
    )
    app = Zayt(settings)

    async with AsyncClient(transport=ASGIWebSocketTransport(app)) as client:
        with pytest.raises(WebSocketDisconnect):
            await aconnect_ws("http://localhost", client).__aenter__()


@pytest.mark.skip
async def test_websocket_new_raising_error():
    settings = Settings(
        default_settings | {"__application__": __package__ + ".application"}
    )
    app = Zayt(settings)

    async with AsyncClient(transport=ASGIWebSocketTransport(app)) as client:
        with pytest.raises(WebSocketDisconnect, match="1011"):
            await aconnect_ws("http://localhost/exception/new", client).__aenter__()


@pytest.mark.skip
async def test_websocket_accepted_raising_error():
    settings = Settings(
        default_settings | {"__application__": __package__ + ".application"}
    )
    app = Zayt(settings)

    async with AsyncClient(transport=ASGIWebSocketTransport(app)) as client:
        with pytest.raises(UnhandledWebSocketEvent):
            async with aconnect_ws("http://localhost/exception/accepted", client) as ws:
                await ws.ping()



import inspect
from types import FunctionType
from typing import Any

from zayt.di.service.model import InjectableType


def _type_name(service):
    if (
        inspect.isclass(service)
        or inspect.isfunction(service)
        or inspect.ismethod(service)
    ):
        return service.__qualname__

    return str(service)


class DependencyInjectionError(Exception):
    pass


class DependencyLoopError(DependencyInjectionError):
    def __init__(
        self, stack: list[tuple[type, str | None]], conflict: tuple[type, str | None]
    ):
        conflict_index = stack.index(conflict)
        last_index = len(stack) - 1

        loop_stack = []
        for i, (dep, name) in enumerate(stack):
            if i == conflict_index:
                indicator = "┌"
            elif i < last_index:
                indicator = "│"
            elif i == last_index:
                indicator = "└"
            else:
                indicator = " "

            dependency = f"{dep.__module__}.{dep.__name__}"
            name = f" ({name})" if name else ""
            loop_stack.append(f"{indicator} {dependency}{name}")

        result = "\n".join(loop_stack)
        super().__init__(f"dependency loop detected: \n{result}")


class ServiceNotFoundError(DependencyInjectionError):
    def __init__(
        self,
        service: type,
        name: str = None,
    ):
        message = f"unable to find service '{_type_name(service)}'"
        if name is not None:
            message += f" with name '{name}'"

        super().__init__(message)


class FactoryMissingReturnTypeError(DependencyInjectionError):
    def __init__(self, factory: FunctionType):
        super().__init__(f"factory '{_type_name(factory)}' is missing return type")


class ServiceAlreadyRegisteredError(DependencyInjectionError):
    def __init__(self, service: type, name: str = None):
        message = f"service '{_type_name(service)}' is already registered"
        if name:
            message += f" with name '{name}'"

        super().__init__(message)


class TypeVarInGenericServiceError(DependencyInjectionError):
    def __init__(self, provided: type):
        super().__init__(f"{_type_name(provided)} has generic types")


class InvalidServiceTypeError(DependencyInjectionError):
    def __init__(self, service: Any):
        super().__init__(f"object of type {type(service)} is not a valid service")


class ServiceWithUntypedDependencyError(DependencyInjectionError):
    def __init__(self, service: InjectableType, param: str):
        super().__init__(
            f"service {service.__module__}.{service.__qualname__}"
            f" must annotate parameter '{param}'"
        )


class InvalidDependencyAnnotationError(DependencyInjectionError):
    def __init__(self, dependency, annotation):
        super().__init__(
            f"dependency '{dependency}' has invalid annotation '{annotation}'"
        )


class IncompatibleDependencyScopeError(DependencyInjectionError):
    def __init__(
        self, service_type: type, dependency: type, dependency_name: str | None
    ):
        super().__init__(
            "incompatible dependency scope in"
            f" '{service_type.__module__}.{service_type.__qualname__}'"
            f" for dependency '{dependency.__module__}.{dependency.__qualname__}'"
        )
        self.service_type = service_type
        self.dependency = dependency
        self.dependency_name = dependency_name


class ScopedServiceWithoutContextError(DependencyInjectionError):
    def __init__(self, service: InjectableType):
        super().__init__(
            f"scoped service {service.__module__}.{service.__qualname__}"
            " requested without context"
        )



from zayt.util.maybe_async import call_maybe_async


async def test_call_sync_function():
    def func():
        return 1

    result = await call_maybe_async(func)
    assert result == 1


async def test_call_sync_function_with_args():
    def func(param: int):
        return param

    result = await call_maybe_async(func, 1)
    assert result == 1


async def test_call_sync_function_with_kwargs():
    def func(param: int):
        return param

    result = await call_maybe_async(func, param=1)
    assert result == 1


async def test_call_sync_method():
    class Klass:
        def func(self):
            return 1

    obj = Klass()

    result = await call_maybe_async(obj.func)
    assert result == 1


async def test_call_sync_method_with_args():
    class Klass:
        def func(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj.func, 1)
    assert result == 1


async def test_call_sync_method_with_kwargs():
    class Klass:
        def func(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj.func, param=1)
    assert result == 1


async def test_call_sync_object():
    class Klass:
        def __call__(self):
            return 1

    obj = Klass()

    result = await call_maybe_async(obj)
    assert result == 1


async def test_call_sync_object_with_args():
    class Klass:
        def __call__(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj, 1)
    assert result == 1


async def test_call_sync_object_with_kwargs():
    class Klass:
        def __call__(self, param: int):
            return param

    obj = Klass()

    result = await call_maybe_async(obj, param=1)
    assert result == 1
